@@ -1,37 +1,14 @@
-/****************************************************************************
-**
-** Copyright (C) 2017 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtWaylandCompositor module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 or (at your option) any later version
-** approved by the KDE Free Qt Foundation. The licenses are as published by
-** the Free Software Foundation and appearing in the file LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2017 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include "qwaylandquickitem.h"
 #include "qwaylandquickitem_p.h"
 #include "qwaylandquicksurface.h"
 #include "qwaylandinputmethodcontrol.h"
 #include "qwaylandtextinput.h"
+#if QT_WAYLAND_TEXT_INPUT_V4_WIP
+#include "qwaylandtextinputv4.h"
+#endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
 #include "qwaylandqttextinputmethod.h"
 #include "qwaylandquickoutput.h"
 #include <QtWaylandCompositor/qwaylandcompositor.h>
@@ -369,7 +346,7 @@ void QWaylandBufferMaterial::setBufferRef(QWaylandQuickItem *surfaceItem, const
         if (auto texture = ref.toOpenGLTexture(plane)) {
             QQuickWindow::CreateTextureOptions opt;
             QWaylandQuickSurface *waylandSurface = qobject_cast<QWaylandQuickSurface *>(surfaceItem->surface());
-            if (waylandSurface != nullptr && waylandSurface->useTextureAlpha())
+            if (waylandSurface != nullptr && waylandSurface->useTextureAlpha() && !waylandSurface->isOpaque())
                 opt |= QQuickWindow::TextureHasAlphaChannel;
             QSGTexture *scenegraphTexture;
             if (ref.bufferFormatEgl() == QWaylandBufferRef::BufferFormatEgl_EXTERNAL_OES) {
@@ -419,7 +396,7 @@ public:
 #if QT_CONFIG(opengl)
                 QQuickWindow::CreateTextureOptions opt;
                 QWaylandQuickSurface *surface = qobject_cast<QWaylandQuickSurface *>(surfaceItem->surface());
-                if (surface && surface->useTextureAlpha()) {
+                if (surface && surface->useTextureAlpha()  && !surface->isOpaque()) {
                     opt |= QQuickWindow::TextureHasAlphaChannel;
                 }
 
@@ -498,12 +475,14 @@ QWaylandQuickItem::~QWaylandQuickItem()
     Q_D(QWaylandQuickItem);
     disconnect(this, &QQuickItem::windowChanged, this, &QWaylandQuickItem::updateWindow);
     QMutexLocker locker(d->mutex);
-    if (d->provider)
+    if (d->provider) {
+        disconnect(d->texProviderConnection);
         d->provider->deleteLater();
+    }
 }
 
 /*!
- * \qmlproperty WaylandCompositor QtWaylandCompositor::WaylandQuickItem::compositor
+ * \qmlproperty WaylandCompositor QtWayland.Compositor::WaylandQuickItem::compositor
  *
  * This property holds the compositor for the surface rendered by this WaylandQuickItem.
  */
@@ -529,7 +508,7 @@ QWaylandView *QWaylandQuickItem::view() const
 }
 
 /*!
- * \qmlproperty WaylandSurface QtWaylandCompositor::WaylandQuickItem::surface
+ * \qmlproperty WaylandSurface QtWayland.Compositor::WaylandQuickItem::surface
  *
  * This property holds the surface rendered by this WaylandQuickItem.
  */
@@ -561,7 +540,7 @@ void QWaylandQuickItem::setSurface(QWaylandSurface *surface)
 }
 
 /*!
- * \qmlproperty enum QtWaylandCompositor::WaylandQuickItem::origin
+ * \qmlproperty enum QtWayland.Compositor::WaylandQuickItem::origin
  *
  * This property holds the origin of the QWaylandQuickItem.
  */
@@ -780,7 +759,7 @@ void QWaylandQuickItem::keyPressEvent(QKeyEvent *event)
 void QWaylandQuickItem::keyReleaseEvent(QKeyEvent *event)
 {
     Q_D(QWaylandQuickItem);
-    if (d->shouldSendInputEvents() && hasFocus()) {
+    if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
         seat->sendFullKeyEvent(event);
     } else {
@@ -874,6 +853,10 @@ void QWaylandQuickItem::handleSubsurfaceAdded(QWaylandSurface *childSurface)
         connect(childSurface, &QWaylandSurface::destroyed, childItem, &QObject::deleteLater);
     } else {
         bool success = QMetaObject::invokeMethod(d->subsurfaceHandler, "handleSubsurfaceAdded", Q_ARG(QWaylandSurface *, childSurface));
+        if (!success)
+            success = QMetaObject::invokeMethod(d->subsurfaceHandler, "handleSubsurfaceAdded",
+                                                Q_ARG(QVariant, QVariant::fromValue(childSurface)));
+
         if (!success)
             qWarning("QWaylandQuickItem: subsurfaceHandler does not implement handleSubsurfaceAdded()");
     }
@@ -914,7 +897,7 @@ void QWaylandQuickItem::handlePlaceBelow(QWaylandSurface *referenceSurface)
 }
 
 /*!
-  \qmlproperty object QtWaylandCompositor::WaylandQuickItem::subsurfaceHandler
+  \qmlproperty object QtWayland.Compositor::WaylandQuickItem::subsurfaceHandler
 
   This property provides a way to override the default subsurface behavior.
 
@@ -926,8 +909,9 @@ void QWaylandQuickItem::handlePlaceBelow(QWaylandSurface *referenceSurface)
   \code
   ShellSurfaceItem {
       subsurfaceHandler: QtObject {
-      function handleSubsurfaceAdded(child) {
-        //create custom surface item, and connect the subsurfacePositionChanged signal
+          function handleSubsurfaceAdded(child) {
+            // create custom surface item, and connect the subsurfacePositionChanged signal
+          }
       }
   }
   \endcode
@@ -974,7 +958,7 @@ void QWaylandQuickItem::setOutput(QWaylandOutput *output)
 }
 
 /*!
- * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::bufferLocked
+ * \qmlproperty bool QtWayland.Compositor::WaylandQuickItem::bufferLocked
  *
  * This property holds whether the item's buffer is currently locked. As long as
  * the buffer is locked, it will not be released and returned to the client.
@@ -1121,13 +1105,24 @@ void QWaylandQuickItem::takeFocus(QWaylandSeat *device)
     }
     target->setKeyboardFocus(surface());
 
-    {
+    qCDebug(qLcWaylandCompositorInputMethods) << Q_FUNC_INFO << " surface:" << surface()
+        << ", client:" << surface()->client()
+        << ", textinputprotocol:" << (int)(surface()->client()->textInputProtocols());
+    if (surface()->client()->textInputProtocols().testFlag(QWaylandClient::TextInputProtocol::TextInputV2)) {
         QWaylandTextInput *textInput = QWaylandTextInput::findIn(target);
         if (textInput)
             textInput->setFocus(surface());
     }
 
-    {
+#if QT_WAYLAND_TEXT_INPUT_V4_WIP
+    if (surface()->client()->textInputProtocols().testFlag(QWaylandClient::TextInputProtocol::TextInputV4)) {
+        QWaylandTextInputV4 *textInputV4 = QWaylandTextInputV4::findIn(target);
+        if (textInputV4)
+            textInputV4->setFocus(surface());
+    }
+#endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
+
+    if (surface()->client()->textInputProtocols().testFlag(QWaylandClient::TextInputProtocol::QtTextInputMethodV1)) {
         QWaylandQtTextInputMethod *textInputMethod = QWaylandQtTextInputMethod::findIn(target);
         if (textInputMethod)
             textInputMethod->setFocus(surface());
@@ -1178,7 +1173,7 @@ void QWaylandQuickItem::updateSize()
 }
 
 /*!
- * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::focusOnClick
+ * \qmlproperty bool QtWayland.Compositor::WaylandQuickItem::focusOnClick
  *
  * This property specifies whether the WaylandQuickItem should take focus when
  * it is clicked or touched.
@@ -1293,7 +1288,7 @@ QVariant QWaylandQuickItem::inputMethodQuery(Qt::InputMethodQuery query, QVarian
 #endif
 
 /*!
-    \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::paintEnabled
+    \qmlproperty bool QtWayland.Compositor::WaylandQuickItem::paintEnabled
 
     Returns true if the item is hidden, though the texture
     is still updated. As opposed to hiding the item by
@@ -1328,7 +1323,7 @@ void QWaylandQuickItem::setPaintEnabled(bool enabled)
 }
 
 /*!
-    \qmlproperty  bool QtWaylandCompositor::WaylandQuickItem::touchEventsEnabled
+    \qmlproperty  bool QtWayland.Compositor::WaylandQuickItem::touchEventsEnabled
 
     This property holds \c true if touch events are forwarded to the client
     surface, \c false otherwise.
@@ -1423,7 +1418,7 @@ void QWaylandQuickItem::updateInputMethod(Qt::InputMethodQueries queries)
 #endif
 
 /*!
- * \qmlsignal void QtWaylandCompositor::WaylandQuickItem::surfaceDestroyed()
+ * \qmlsignal void QtWayland.Compositor::WaylandQuickItem::surfaceDestroyed()
  *
  * This signal is emitted when the client has destroyed the \c wl_surface associated
  * with the WaylandQuickItem. The handler for this signal is expected to either destroy the
@@ -1490,8 +1485,24 @@ QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDat
             d->newTexture = true;
         }
 
-        if (!d->provider)
+        if (!d->provider) {
             d->provider = new QWaylandSurfaceTextureProvider();
+            if (compositor()) {
+                d->texProviderConnection =
+                    QObject::connect(
+                            compositor(),
+                            &QObject::destroyed,
+                            this,
+                            [this](QObject*) {
+                                    auto *itemPriv = QWaylandQuickItemPrivate::get(this);
+                                    if (itemPriv->provider) {
+                                        itemPriv->provider->deleteLater();
+                                        itemPriv->provider = nullptr;
+                                    }
+                                    disconnect(itemPriv->texProviderConnection); }
+                    );
+            }
+        }
 
         if (d->newTexture) {
             d->newTexture = false;
@@ -1593,20 +1604,34 @@ void QWaylandQuickItem::setInputEventsEnabled(bool enabled)
 
 void QWaylandQuickItem::lower()
 {
-    QQuickItem *parent = parentItem();
+    Q_D(QWaylandQuickItem);
+    d->lower();
+}
+
+void QWaylandQuickItemPrivate::lower()
+{
+    Q_Q(QWaylandQuickItem);
+    QQuickItem *parent = q->parentItem();
     Q_ASSERT(parent);
-    QQuickItem *bottom = parent->childItems().first();
-    if (this != bottom)
-        stackBefore(bottom);
+    QQuickItem *bottom = parent->childItems().constFirst();
+    if (q != bottom)
+        q->stackBefore(bottom);
 }
 
 void QWaylandQuickItem::raise()
 {
-    QQuickItem *parent = parentItem();
+    Q_D(QWaylandQuickItem);
+    d->raise();
+}
+
+void QWaylandQuickItemPrivate::raise()
+{
+    Q_Q(QWaylandQuickItem);
+    QQuickItem *parent = q->parentItem();
     Q_ASSERT(parent);
-    QQuickItem *top = parent->childItems().last();
-    if (this != top)
-        stackAfter(top);
+    QQuickItem *top = parent->childItems().constLast();
+    if (q != top)
+        q->stackAfter(top);
 }
 
 void QWaylandQuickItem::sendMouseMoveEvent(const QPointF &position, QWaylandSeat *seat)
